//
//  kmpg
// 
//  Copyright (C) 1997 Bjarni Ivarsson
//  email: Bjarni.Ivarsson@swipnet.se
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//
//  A playlist manager.
//

#include <stdlib.h>
#include <iostream.h>

#include <qdir.h>
#include <qfile.h>
#include <qfileinf.h>
#include <qtstream.h>

#include "playlistmgr.h"
#include "fileinfo.h"

PlaylistMgr::PlaylistMgr()
{
   _list.setAutoDelete( TRUE ); // delete items when they are removed
   _playmode = Random;
   _repeat = Once;
   _changed = FALSE;
   _lastTrack = -1;
}

PlaylistMgr::~PlaylistMgr()
{
}

void PlaylistMgr::saveProperties( KConfig* config )
{
   QString filename = kapp->localkdedir() + "/share/apps/kmpg/plist.m3u";
   KURL url( filename );
   QString played = "";
   Track *track;

   if( _list.count() > 0 ) 
   {
      config->writeEntry( "Playlist", filename );
      save( url ); // Save the playlist to a temp file

      // Now save the status of the individual tracks
      for( track=_list.first(); track != 0; track=_list.next() )
	 played += track->isPlayed()?"x":"o";
      config->writeEntry( "Played", played );
   }
}

void PlaylistMgr::readProperties( KConfig* config )
{
   QString filename, played = "";
   Track *track;
   int i;
   
   filename = config->readEntry( "Playlist" );
   if( !filename.isNull() )
   {
      load( KURL(filename) );
      
      // Get track status
      played = config->readEntry( "Played" );
      for( track=_list.first(), i=1; track != 0; track=_list.next(), i++ )
      {
	 if( played.left( 1 ) == "x" )
	 {
	    track->setPlayed( TRUE );
	    emit( sigSetPlayed( i, TRUE ) );
	 }
	 played = played.mid( 2, played.length() - 1 );
      }
   }
}

// Load a new playlist
void PlaylistMgr::load( KURL url )
{
   QFile file;
   QTextStream stream;
   QString string, filename;
   
   filename = url.path();
   file.setName( filename );
   file.open( IO_ReadOnly );
   stream.setDevice( &file );
   clear();
   string = stream.readLine();
   while( !stream.eof() )
   {
      insert( KURL(string.data()) );
      string = stream.readLine();
   }
   file.close();
   _changed = FALSE;
   emit( sigChanged(_changed) );
}

// Add playlist to current list
void PlaylistMgr::add( KURL url )
{
   QFile file;
   QTextStream stream;
   QString string, filename;
   
   filename = url.path();
   file.setName( filename );
   file.open( IO_ReadOnly );
   stream.setDevice( &file );
   string = stream.readLine();
   while( !stream.eof() )
   {
      insert( KURL(string.data()) );
      string = stream.readLine();
   }
   file.close();
   _changed = TRUE;
   emit( sigChanged(_changed) );
}

// Save the playlist
void PlaylistMgr::save( KURL url )
{
   QFile file;
   QTextStream stream;
   QString filename;
   Track *track;
   
   filename = url.path();
   file.setName( filename );
   file.open( IO_WriteOnly );
   stream.setDevice( &file );
   for( track=_list.first(); track != 0; track=_list.next() )
   {
      stream << track->getUrl().path() << endl;
   }
   file.close();
   _changed = FALSE;
   emit( sigChanged(_changed) );
}

void PlaylistMgr::clear()
{
   _list.clear();
   _lastTrack = -1;
   _track = 0;
   setList(FALSE);
   emit( sigClear() );
   emit( sigChanged(_changed) );
}

// Insert a track into the playlist
// Can handle adding directories correctly
void PlaylistMgr::insert( KURL url )
{
   QFileInfo *info;

   info = new QFileInfo( url.path() ); // For file: protocol
   if( info->isDir() )
   {
      QDir dir( url.path(), "*", 
		QDir::DirsFirst ,
		QDir::Dirs|QDir::Files|QDir::NoSymLinks );
      int item;

      cerr << "Directory: " << url.path() << endl;
      for( item=0; item<dir.count(); item++ ) 
      {
	 if( *(dir[item])=='.' ) // Ignore hidden files
	    continue;
	 cerr << "[" << dir[item] << "]";
	 insert( KURL(url, dir[item]) );
      }
   }
   else
   {
      Track *track;
   
      cerr << "File: " << url.path() << endl;
      Fileinfo info(url.path());
      if ( info.isOk() ) 
      { 
         track = new Track( url );
	 track->setLength( (int)(info.getMinNumberOfFrames()*info.getMsPerFrame()/1000+0.5) );
	 track->setPlayed(FALSE);
         _list.append( track );
         emit( sigInsert( _list.count()-1, url.path() ) );
//	 emit( sigSetPlayed( _list.count()-1, FALSE ));
         _changed = TRUE;
         emit( sigChanged(_changed) );
      }
   }
   delete info;
}

// Insert a track and play it
void PlaylistMgr::insertPlay( KURL url )
{
   insert( url );
   playTrack( _list.count()-1 );
}

// Remove track at a specific index
void PlaylistMgr::remove( int index )
{
   Track *track;
   
   track = _list.at( index );
   if ( index == (getTrack() - 1)  )
      setList(FALSE);
   if ( index < (getTrack() - 1) )
      _track--;
   _list.removeRef( track );
   emit( sigRemove( index ) );
   if( _lastTrack > index )
      _lastTrack--;
   _changed = TRUE;
   emit( sigChanged(_changed) );
}

void PlaylistMgr::moveUp( int index )
{
   _changed = TRUE;
   emit( sigChanged(_changed) );
}

void PlaylistMgr::moveDown( int index )
{
   _changed = TRUE;
   emit( sigChanged(_changed) );
}

// Play a track at a specific index
void PlaylistMgr::playTrack( int index )
{
   cerr << "PlayTrack called with index=" << index << "\n" << endl;
   Track *track;
   _track = index;
   if( index<0 || index>=_list.count() )
      return;

   track = _list.at( index );
   track->setPlayed( TRUE );
   emit( sigSetPlayed( index, TRUE ) );
   emit( sigTrackSelected( track->getUrl(), TRUE ) );
   emit( sigTrackSelected( index ) );
   _lastTrack = index;
   setList(TRUE);
}

// Play the next track, taking into account playmode (random/sequential)
// and wether to use repeat play
void PlaylistMgr::playNext()
{
   cout << "PlayNext() called\n" << endl;
   int i=_lastTrack;
   
   if( _list.count() == 0 )
      return;

   if ( _repeat == One ) {
      playTrack(_lastTrack);
      return;
   }

   if( areAllPlayed() )
      if( _repeat == All )
      {
	 setAllPlayed( FALSE );
	 if( _playmode==Random )
	 {
	    doRandom();
	    return;
	 }
	 else
	 {
	    playTrack(0);
	    return;
	 }
      }
      else
	 return;
   
   if( _playmode==Random )
   {
      doRandom();
      return;
   }
  
   if ( i < 0 ) i = 0;
   while( _list.at(i)->isPlayed() )
   {
      i++;
      if( i>=_list.count() )
	 i=0;
   }
   playTrack( i );
}

// Play the previous track
void PlaylistMgr::playPrev()
{
   if( _list.count() == 0 )
      return;

   playNext();
}

// Check if all the tracks have been played
bool PlaylistMgr::areAllPlayed()
{
   bool ret = true;
   Track *track;
   
   for( track=_list.first(); track != 0; track=_list.next() )
      if( !track->isPlayed() )
	 ret = false;

   return ret;
}

// Set the playstatus of all tracks
void PlaylistMgr::setAllPlayed( bool pl )
{
   Track *track;
   
   for( track=_list.first(); track != 0; track=_list.next() )
      track->setPlayed( pl );
   emit( sigSetAllPlayed( pl ) );
}

// Play a track at random
void PlaylistMgr::doRandom()
{
   int num = 0, random;
   Track *track;
   
   for( track=_list.first(); track != 0; track=_list.next() )
      if( !track->isPlayed() )
	 num++;
   // returns a number between 1 and num
   random = 1 + (int)( ((float)num)*rand()/(RAND_MAX+1.0) );

   for( track=_list.first(), num=0; 
	track != 0 && random>0; 
	track=_list.next(), num++ )
      if( !track->isPlayed() )
	 random--;
   
   playTrack( num - 1 );
}

int PlaylistMgr::getPlayed()
{
   int i = 0;
   Track *t;
   if ( !isList() || getTracks() <= 0 )
      return 0;
   for ( t = _list.first(); t != 0; t = _list.next() )
      if ( t->isPlayed() ) i++;
   return i;
}

int PlaylistMgr::getListTime()
{
   int i = 0;
   Track *t;
   if ( !isList() || getTracks() <= 0 )
      return 0;
   for ( t = _list.first(); t != 0; t = _list.next() )
      i += t->Length();
   return i;
}

int PlaylistMgr::getPlayTime()
{
   int i = 0;
   Track *t;
   if ( !isList() || getTracks() <= 0 )
      return 0;
   for ( t = _list.first(); t != 0; t = _list.next() )
      if ( t->isPlayed() )
         i += t->Length();
      t = _list.at(_track);
      i -= t->Length();
   return i;
}

